--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- This testbench tests the different type of DDR controllers: 
--
--                        - uphy_4g_800_master
--                        - uphy_4g_800_slave
--
-- The DUT can be selected, using the c_tech_ddr constants. 
--
-- Testbench is selftesting: 
--
-- > run -all 
--

LIBRARY IEEE, technology_lib, tech_ddr_lib, common_lib, dp_lib, diagnostics_lib;                   
USE IEEE.STD_LOGIC_1164.ALL;    
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;

ENTITY tb_io_ddr IS  
END ENTITY tb_io_ddr;

ARCHITECTURE str of tb_io_ddr IS

  CONSTANT c_ctlr_ref_clk_period : TIME := 5 ns;  -- 200 MHz
  CONSTANT c_dp_clk_period       : TIME := 5 ns;  -- 200 MHz

  CONSTANT c_tech_ddr         : t_c_tech_ddr := c_tech_ddr3_4g_800m_master;
  
--  CONSTANT c_ddr_addr_hi      : t_tech_ddr_addr := ((OTHERS=>'0'),
--                                                    (OTHERS=>'0'),
--                                                    TO_UVEC(                                         3, c_tech_ddr_max.a_row_w),
--                                                    TO_UVEC(2**c_tech_ddr_4g.a_col_w-c_tech_ddr_4g.rsl, c_tech_ddr_max.a_col_w));
  CONSTANT c_ddr_addr_hi      : t_tech_ddr_addr := ((OTHERS=>'0'),
                                                    (OTHERS=>'0'),
                                                    TO_UVEC(                                         3, c_tech_ddr_max.a_row_w),
                                                    TO_UVEC(2**c_tech_ddr_4g.a_col_w-1, c_tech_ddr_max.a_col_w));
                                                    
  CONSTANT c_ctlr_data_w      : NATURAL := func_tech_ddr_ctlr_data_w(c_tech_ddr);
  CONSTANT c_dp_data_w        : NATURAL := 32;
  CONSTANT c_dp_factor        : NATURAL := c_ctlr_data_w/c_dp_data_w;
 
  SIGNAL ctlr_ref_clk         : STD_LOGIC := '0';
  SIGNAL ctlr_ref_rst         : STD_LOGIC;
  SIGNAL tb_end               : STD_LOGIC := '0';
  SIGNAL dp_clk               : STD_LOGIC := '0';
  SIGNAL dp_rst               : STD_LOGIC;

  SIGNAL ctlr_rdy             : STD_LOGIC;
  SIGNAL ctlr_init_done       : STD_LOGIC;

  SIGNAL dvr_start_addr       : t_tech_ddr_addr; 
  SIGNAL dvr_end_addr         : t_tech_ddr_addr;

  SIGNAL dvr_en               : STD_LOGIC;
  SIGNAL dvr_wr_not_rd        : STD_LOGIC;
  SIGNAL dvr_done             : STD_LOGIC;

  SIGNAL wr_siso              : t_dp_siso;
  SIGNAL wr_sosi              : t_dp_sosi;

  SIGNAL rd_siso              : t_dp_siso;
  SIGNAL rd_sosi              : t_dp_sosi;

  SIGNAL src_diag_en          : STD_LOGIC;
  SIGNAL src_val_cnt          : STD_LOGIC_VECTOR(31 DOWNTO 0);

  SIGNAL snk_diag_en          : STD_LOGIC;
  SIGNAL snk_diag_res         : STD_LOGIC;
  SIGNAL snk_diag_res_val     : STD_LOGIC;
  SIGNAL snk_val_cnt          : STD_LOGIC_VECTOR(31 DOWNTO 0);
  SIGNAL expected_cnt         : NATURAL;
  
  SIGNAL phy_in               : t_tech_ddr_phy_in;
  SIGNAL phy_io               : t_tech_ddr_phy_io;
  SIGNAL phy_ou               : t_tech_ddr_phy_ou;
  
BEGIN
 
  ctlr_ref_clk <= NOT ctlr_ref_clk OR tb_end AFTER c_ctlr_ref_clk_period/2; 
  ctlr_ref_rst <= '1', '0' AFTER 100 ns;
  
  dp_clk   <= NOT dp_clk OR tb_end AFTER c_dp_clk_period/2; 
  dp_rst   <= '1', '0' AFTER 100 ns;

  dvr_start_addr <= c_tech_ddr_addr_lo;
  dvr_end_addr   <= c_ddr_addr_hi;

  p_stimuli : PROCESS
  BEGIN
    tb_end        <= '0';
    dvr_en        <= '0';
    src_diag_en   <= '0';
    dvr_wr_not_rd <= '0';
    snk_diag_en   <= '0';

    proc_common_wait_until_high(dp_clk, ctlr_init_done);
    proc_common_wait_some_cycles(dp_clk, 2);  -- Give the driver FSM a cycle to go into idle mode
    
    -- START diagnostics source for write and sink for verify read
    src_diag_en   <= '1';
    snk_diag_en   <= '1';
    
    -- START WRITE
    dvr_wr_not_rd <= '1';
    dvr_en        <= '1';
    proc_common_wait_some_cycles(dp_clk, 1);
    dvr_en        <= '0'; 
    
    -- WRITE DONE  
    proc_common_wait_some_cycles(dp_clk, 10);
    proc_common_wait_until_high(dp_clk, dvr_done);

    -- START READ
    dvr_wr_not_rd <= '0';
    dvr_en        <= '1';
    proc_common_wait_some_cycles(dp_clk, 1);
    dvr_en        <= '0'; 

    -- READ DONE
    proc_common_wait_some_cycles(dp_clk, 10);
    proc_common_wait_until_high(dp_clk, dvr_done);

    expected_cnt <= 1024*c_dp_factor;
    proc_common_wait_some_cycles(dp_clk, 500*c_dp_factor);  -- 'Done' means all requests are posted. Wait for the last read data to arrive.
  
    ASSERT snk_diag_res_val = '1'               REPORT "[ERROR] DIAG_RES INVALID!"               SEVERITY FAILURE;
    ASSERT snk_diag_res = '0'                   REPORT "[ERROR] NON-ZERO DIAG_RES!"              SEVERITY FAILURE;
    ASSERT UNSIGNED(snk_val_cnt) = expected_cnt REPORT "[ERROR] Unexpected number of read data!" SEVERITY FAILURE;
    
    REPORT "[OK] Test passed." SEVERITY NOTE;
    tb_end <= '1';

    WAIT;
  END PROCESS;

  u_diagnostics: ENTITY diagnostics_lib.diagnostics 
  GENERIC MAP (
    g_dat_w             => c_dp_data_w,
    g_nof_streams       => 1
     ) 
  PORT MAP (
    rst                 => dp_rst,
    clk                 => dp_clk,

    snk_out_arr(0)      => rd_siso,
    snk_in_arr(0)       => rd_sosi,
    snk_diag_en(0)      => snk_diag_en,
    snk_diag_md(0)      => '1',
    snk_diag_res(0)     => snk_diag_res,
    snk_diag_res_val(0) => snk_diag_res_val,
    snk_val_cnt(0)      => snk_val_cnt,

    src_out_arr(0)      => wr_sosi,
    src_in_arr(0)       => wr_siso,
    src_diag_en(0)      => src_diag_en,
    src_diag_md(0)      => '1',
    src_val_cnt(0)      => src_val_cnt
  );

  u_io_ddr: ENTITY work.io_ddr
  GENERIC MAP(
    g_technology       => c_tech_select_default,
    g_tech_ddr         => c_tech_ddr,
    g_wr_data_w        => c_dp_data_w,
    g_rd_data_w        => c_dp_data_w
  )                      
  PORT MAP (                  
    ctlr_ref_clk       => ctlr_ref_clk,
    ctlr_ref_rst       => ctlr_ref_rst,
                                     
    ctlr_init_done     => ctlr_init_done,
    ctlr_rdy           => ctlr_rdy,

    dvr_start_addr     => dvr_start_addr,
    dvr_end_addr       => dvr_end_addr,
    dvr_en             => dvr_en,
    dvr_wr_not_rd      => dvr_wr_not_rd,
    dvr_done           => dvr_done,
 
    wr_clk             => dp_clk,
    wr_rst             => dp_rst,

    wr_sosi            => wr_sosi, 
    wr_siso            => wr_siso,
  
    rd_sosi            => rd_sosi,
    rd_siso            => rd_siso,

    rd_clk             => dp_clk,
    rd_rst             => dp_rst,

    phy_ou             => phy_ou,
    phy_io             => phy_io,
    phy_in             => phy_in
  );
 
  u_tech_ddr_memory_model : ENTITY tech_ddr_lib.tech_ddr_memory_model
  GENERIC MAP (
    g_tech_ddr => c_tech_ddr
  )
  PORT MAP (
    mem_in => phy_ou,
    mem_io => phy_io
  );
    
END ARCHITECTURE str;


